from django.apps import AppConfig
from django.conf import settings
from django.template import engines

# fmt: off
class CottonConfig(AppConfig):
    name = "django_cotton"
    verbose_name = "Cotton"

    def ready(self):
        self._modify_templates_settings()
        self._add_builtin_template_tag()

    def _modify_templates_settings(self):
        modified = False
        for template in settings.TEMPLATES:
            if not template.get("APP_DIRS", True):
                # If APP_DIRS is explicitly set to False, we assume the user
                # has a custom setup and do not modify the settings and provide tutorial instead.
                continue

            # Example modification: Set APP_DIRS to False
            template["APP_DIRS"] = False

            # Add your custom template loader
            if "OPTIONS" not in template:
                template["OPTIONS"] = {}
            if "loaders" not in template["OPTIONS"]:
                template["OPTIONS"]["loaders"] = []

            # Add django_cotton loader
            if "django_cotton.cotton_loader.Loader" not in template["OPTIONS"]["loaders"]:
                template["OPTIONS"]["loaders"].insert(
                    0, "django_cotton.cotton_loader.Loader"
                )

            # Ensure default loaders are present, then add your custom loader
            if "django.template.loaders.filesystem.Loader" not in template["OPTIONS"]["loaders"]:
                template["OPTIONS"]["loaders"].append(
                    "django.template.loaders.filesystem.Loader"
                )
            if "django.template.loaders.app_directories.Loader" not in template["OPTIONS"]["loaders"]:
                template["OPTIONS"]["loaders"].append(
                    "django.template.loaders.app_directories.Loader"
                )

            # Specify TEMPLATE_DIRS if necessary
            # template['DIRS'] += ['path/to/your/templates']

            modified = True

        if modified:
            print("TEMPLATES setting modified by Django Cotton.")

    def _add_builtin_template_tag(self):
        """Add a custom template tag to the built-ins."""
        builtins = engines["django"].engine.builtins
        custom_tag_lib = "django_cotton.templatetags.cotton"
        if custom_tag_lib not in builtins:
            builtins.append(custom_tag_lib)



from django import template
from django.template import Node
from django.template.loader import render_to_string


def cotton_component(parser, token):
    bits = token.split_contents()
    tag_name = bits[0]
    template_path = bits[1]
    component_key = bits[2]

    kwargs = {}
    for bit in bits[3:]:
        key, value = bit.split("=")
        kwargs[key] = str(value.strip("'\""))

    nodelist = parser.parse(("end_cotton_component",))
    parser.delete_first_token()

    return CottonComponentNode(nodelist, template_path, component_key, kwargs)


class CottonComponentNode(Node):
    def __init__(self, nodelist, template_path, component_key, kwargs):
        self.nodelist = nodelist
        self.template_path = template_path
        self.component_key = component_key
        self.kwargs = kwargs

    def render(self, context):
        local_context = context.flatten()

        attrs = {}

        # Process attrs + props
        for key, value in self.kwargs.items():
            try:
                resolved_value = template.Variable(value).resolve(context)
                attrs.update({key: resolved_value})
            except TypeError:
                attrs.update({key: value})
            except template.VariableDoesNotExist:
                attrs.update({key: value})

        # Add the remainder as the default slot
        rendered = self.nodelist.render(context)
        local_context.update({"slot": rendered})

        slots = context.get("cotton_slots", {})
        component_slots = slots.get(self.component_key, {})

        local_context.update(component_slots)
        local_context.update(attrs)

        local_context.update({"attrs_dict": attrs})

        rendered = render_to_string(self.template_path, local_context)

        # Now reset the component slots in context
        slots[self.component_key] = {}
        context.update({"cotton_slots": slots})

        return rendered



from django import template
from django.utils.safestring import mark_safe


def cotton_slot(parser, token):
    try:
        tag_name, slot_name, component_key = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError("incomplete c-slot %r" % token.contents)

    nodelist = parser.parse(("end_cotton_slot",))
    parser.delete_first_token()
    return CottonSlotNode(slot_name, nodelist, component_key)


class CottonSlotNode(template.Node):
    def __init__(self, slot_name, nodelist, component_key):
        self.slot_name = slot_name
        self.nodelist = nodelist
        self.component_key = component_key

    def render(self, context):
        # Add the rendered content to the context.
        if "cotton_slots" not in context:
            context.update({"cotton_slots": {}})

        # context["cotton_slots"][self.slot_name] = mark_safe(output)

        output = self.nodelist.render(context)

        # with context.push():
        # Temporarily store the slot's content in the new layer
        # if "cotton_slots" not in context:
        #     context["cotton_slots"] = {}

        if self.component_key not in context["cotton_slots"]:
            context["cotton_slots"][self.component_key] = {}

        # if self.slot_name not in context["cotton_slots"][self.component_key]:
        #     context["cotton_slots"][self.component_key][self.slot_name] = mark_safe(output)
        context["cotton_slots"][self.component_key][self.slot_name] = mark_safe(output)

        # context.push()

        # todo add scoping by component
        # context["cotton_slots"][self.component_key][self.slot_name] = mark_safe(output)

        return ""



from django.views.generic import TemplateView
from . import views
from django.urls import path

app_name = "django_cotton"


class NamedSlotInLoop(TemplateView):
    template_name = "named_slot_in_loop.cotton.html"

    def get_context_data(self, **kwargs):
        return {
            "items": [
                {"name": "Item 1"},
                {"name": "Item 2"},
                {"name": "Item 3"},
            ]
        }


urlpatterns = [
    path("parent", TemplateView.as_view(template_name="parent_test.cotton.html")),
    path("child", TemplateView.as_view(template_name="child_test.cotton.html")),
    path(
        "self-closing",
        TemplateView.as_view(template_name="self_closing_test.cotton.html"),
    ),
    path("include", TemplateView.as_view(template_name="cotton_include.cotton.html")),
    path("playground", TemplateView.as_view(template_name="playground.cotton.html")),
    path("tag", TemplateView.as_view(template_name="tag.cotton.html")),
    path("named-slot-in-loop", NamedSlotInLoop.as_view()),
    path("test/compiled-cotton", views.compiled_cotton_test_view),
    path("test/cotton", views.cotton_test_view),
    path("test/native-extends", views.native_extends_test_view),
    path("test/native-include", views.native_include_test_view),
    path("attribute-merging", views.attribute_merging_test_view),
    path("attribute-passing", views.attribute_passing_test_view),
    path("django-syntax-decoding", views.django_syntax_decoding_test_view),
    path(
        "string-with-spaces",
        TemplateView.as_view(template_name="string_with_spaces.cotton.html"),
    ),
    path("props-test", TemplateView.as_view(template_name="props_test.cotton.html")),
]



import os
import re
import hashlib
import warnings

from django.template.loaders.base import Loader as BaseLoader
from django.core.exceptions import SuspiciousFileOperation
from django.template import TemplateDoesNotExist
from django.utils._os import safe_join
from django.template import Template
from django.core.cache import cache
from django.template import Origin
from django.conf import settings

from bs4 import BeautifulSoup, MarkupResemblesLocatorWarning


warnings.filterwarnings("ignore", category=MarkupResemblesLocatorWarning)


class Loader(BaseLoader):
    is_usable = True

    def __init__(self, engine, dirs=None):
        super().__init__(engine)
        self.cache_handler = CottonTemplateCacheHandler()
        self.dirs = dirs
        self.django_syntax_placeholders = []

    def get_template_from_string(self, template_string):
        """Create and return a Template object from a string. Used primarily for testing."""
        return Template(template_string, engine=self.engine)

    def get_template(self, template_name, skip=None):
        """
        Call self.get_template_sources() and re
        turn a Template object for
        the first template matching template_name. If skip is provided, ignore
        template origins in skip. This is used to avoid recursion during
        template extending.
        """

        tried = []

        for origin in self.get_template_sources(template_name):
            if skip is not None and origin in skip:
                tried.append((origin, "Skipped to avoid recursion"))
                continue

            try:
                contents = self.get_contents(origin)
            except TemplateDoesNotExist:
                tried.append((origin, "Source does not exist"))
                continue
            else:
                return Template(
                    contents,
                    origin,
                    origin.template_name,
                    self.engine,
                )

        raise TemplateDoesNotExist(template_name, tried=tried)

    def get_contents(self, origin):
        # check if file exists, whilst getting the mtime for cache key
        try:
            mtime = os.path.getmtime(origin.name)
        except FileNotFoundError:
            raise TemplateDoesNotExist(origin)

        # check and return cached template
        cache_key = self.cache_handler.get_cache_key(origin.template_name, mtime)
        cached_content = self.cache_handler.get_cached_template(cache_key)
        if cached_content is not None:
            return cached_content

        # If not cached, process the template
        template_string = self._get_template_string(origin.name)

        # We need to provide a key to the current view or component (in this case, view) so that we can namespace
        # slot data, preventing bleeding and ensure component's clear only data in the context applicable to itself
        # in this case, we're top level, likely in a view so we use the view template name as the key
        component_key = (
            origin.template_name.lstrip("cotton/")
            .rstrip(".cotton.html")
            .replace("/", ".")
        )

        compiled_template = self._compile_template_from_string(
            template_string, component_key
        )

        # Cache the processed template
        self.cache_handler.cache_template(cache_key, compiled_template)

        return compiled_template

    def _replace_syntax_with_placeholders(self, content):
        """# replace {% ... %} and {{ ... }} with placeholders so they dont get touched
        or encoded by bs4. Store them to later switch them back in after transformation.
        """
        self.django_syntax_placeholders = []

        # First handle cotton_verbatim blocks, this is designed to preserve and display cotton syntax,
        # akin to the verbatim tag in Django.
        def replace_cotton_verbatim(match):
            inner_content = match.group(
                1
            )  # Get the inner content without the cotton_verbatim tags
            self.django_syntax_placeholders.append(inner_content)
            return f"__django_syntax__{len(self.django_syntax_placeholders)}__"

        # Replace cotton_verbatim blocks, capturing inner content
        content = re.sub(
            r"\{% cotton_verbatim %\}(.*?)\{% endcotton_verbatim %\}",
            replace_cotton_verbatim,
            content,
            flags=re.DOTALL,
        )

        content = re.sub(
            r"\{%.*?%\}",
            lambda x: self.django_syntax_placeholders.append(x.group(0))
            or f"__django_syntax__{len(self.django_syntax_placeholders)}__",
            content,
        )
        content = re.sub(
            r"\{\{.*?\}\}",
            lambda x: self.django_syntax_placeholders.append(x.group(0))
            or f"__django_syntax__{len(self.django_syntax_placeholders)}__",
            content,
        )

        return content

    def _replace_placeholders_with_syntax(self, content):
        """After modifying the content, replace the placeholders with the django template tags and variables."""
        for i, placeholder in enumerate(self.django_syntax_placeholders, 1):
            content = content.replace(f"__django_syntax__{i}__", placeholder)

        return content

    def _get_template_string(self, template_name):
        try:
            with open(template_name, "r") as f:
                content = f.read()
        except FileNotFoundError:
            raise TemplateDoesNotExist(template_name)

        return content

    def _compile_template_from_string(self, content, component_key):
        content = self._replace_syntax_with_placeholders(content)
        content = self._compile_cotton_to_django(content, component_key)
        content = self._replace_placeholders_with_syntax(content)
        content = self._revert_bs4_attribute_empty_attribute_fixing(content)

        return content

    def _revert_bs4_attribute_empty_attribute_fixing(self, contents):
        """Django's template parser adds ="" to empty attribute-like parts in any html-like node, i.e. <div {{ something }}> gets
        compiled to <div {{ something }}=""> Then if 'something' is holding attributes sets, the last attribute value is
        not quoted. i.e. model=test not model="test"."""
        cleaned_content = re.sub('}}=""', "}}", contents)
        return cleaned_content

    def get_dirs(self):
        return self.dirs if self.dirs is not None else self.engine.dirs

    def get_template_sources(self, template_name):
        """Return an Origin object pointing to an absolute path in each directory
        in template_dirs. For security reasons, if a path doesn't lie inside
        one of the template_dirs it is excluded from the result set."""
        if template_name.endswith(".cotton.html"):
            for template_dir in self.get_dirs():
                try:
                    name = safe_join(template_dir, template_name)
                except SuspiciousFileOperation:
                    # The joined path was located outside of this template_dir
                    # (it might be inside another one, so this isn't fatal).
                    continue

                yield Origin(
                    name=name,
                    template_name=template_name,
                    loader=self,
                )

    def _compile_cotton_to_django(self, html_content, component_key):
        """Convert cotton <c-* syntax to {%."""
        soup = BeautifulSoup(html_content, "html.parser")

        soup = self._wrap_with_cotton_props_frame(soup)
        self._transform_components(soup, component_key)

        return str(soup)

    def _transform_prop_tags(self, soup):
        c_props = soup.find_all("c-props")

        for tag in c_props:
            # Build the cotton_props tag string
            props_list = []
            for prop, value in tag.attrs.items():
                if value is None:
                    props_list.append(prop)
                else:
                    props_list.append(f'{prop}="{value}"')

            cotton_props_str = "{% cotton_props " + " ".join(props_list) + " %}"

            # Replace the <c-props> tag with the cotton_props string
            tag.replace_with(cotton_props_str)

        return soup

    def _wrap_with_cotton_props_frame(self, soup):
        """Wrap content with {% cotton_props_frame %} to be able to govern props and attributes."""
        props_with_defaults = []
        c_props = soup.find("c-props")

        # parse c-props tag to extract properties and defaults
        if c_props:
            props_with_defaults = []
            for prop, value in c_props.attrs.items():
                if value is None:
                    props_with_defaults.append(f"{prop}={prop}")
                else:
                    # Assuming value is already a string that represents the default value
                    props_with_defaults.append(f'{prop}={prop}|default:"{value}"')

            c_props.decompose()

        # Construct the {% with %} opening tag
        opening = "{% cotton_props_frame " + " ".join(props_with_defaults) + " %}"
        closing = "{% endcotton_props_frame %}"

        # Convert the remaining soup back to a string and wrap it within {% with %} block
        wrapped_content = opening + str(soup).strip() + closing

        # Since we can't replace the soup object itself, we create new soup instead
        new_soup = BeautifulSoup(wrapped_content, "html.parser")

        return new_soup

    def _transform_named_slot(self, slot_tag, component_key):
        """Replace <c-slot> tags with the {% cotton_slot %} template tag"""
        # for c_slot in soup.find_all("c-slot"):
        slot_name = slot_tag.get("name", "").strip()
        inner_html = "".join(str(content) for content in slot_tag.contents)

        # Check and process any components in the slot content

        slot_soup = BeautifulSoup(inner_html, "html.parser")
        self._transform_components(slot_soup, component_key)

        cotton_slot_tag = f"{{% cotton_slot {slot_name} {component_key} %}}{str(slot_soup)}{{% end_cotton_slot %}}"

        slot_tag.replace_with(BeautifulSoup(cotton_slot_tag, "html.parser"))

    def _transform_components(self, soup, component_key):
        """Replace <c-[component path]> tags with the {% cotton_component %} template tag"""
        for tag in soup.find_all(re.compile("^c-"), recursive=True):
            if tag.name == "c-slot":
                self._transform_named_slot(tag, component_key)

                continue

            component_name = tag.name[2:]

            # Convert dot notation to path structure and replace hyphens with underscores
            path = component_name.replace(".", "/").replace("-", "_")

            # Construct the opening tag
            opening_tag = f"{{% cotton_component {'cotton/{}.cotton.html'.format(path)} {component_name} "
            for attr, value in tag.attrs.items():
                if attr == "class":
                    value = " ".join(value)
                opening_tag += ' {}="{}"'.format(attr, value)
            opening_tag += " %}"

            # Construct the closing tag
            closing_tag = "{% end_cotton_component %}"

            if tag.contents:
                component_key = component_name

                tag_soup = BeautifulSoup(tag.decode_contents(), "html.parser")
                self._transform_components(tag_soup, component_key)

                # Create new content with opening tag, tag content, and closing tag
                new_content = opening_tag + str(tag_soup) + closing_tag

            else:
                # Create new content with opening tag and closing tag
                new_content = opening_tag + closing_tag

            # Replace the original tag with the new content
            new_soup = BeautifulSoup(new_content, "html.parser")
            tag.replace_with(new_soup)

        return soup


class CottonTemplateCacheHandler:
    """Handles caching of cotton templates so the html parsing is only done on first load of each view or component."""

    def __init__(self):
        self.enabled = getattr(settings, "TEMPLATE_CACHING_ENABLED", True)

    def get_cache_key(self, template_name, mtime):
        template_hash = hashlib.sha256(template_name.encode()).hexdigest()
        return f"cotton_cache_{template_hash}_{mtime}"

    def get_cached_template(self, cache_key):
        if not self.enabled:
            return None
        return cache.get(cache_key)

    def cache_template(self, cache_key, content, timeout=None):
        if self.enabled:
            cache.set(cache_key, content, timeout=timeout)



from django import template

register = template.Library()


def cotton_props(parser, token):
    # Split the token to get variable assignments
    parts = token.split_contents()
    cotton_props = {}
    for part in parts[1:]:
        key, value = part.split("=")
        cotton_props[key] = value

    return CottonPropNode(cotton_props)


class CottonPropNode(template.Node):
    def __init__(self, cotton_props):
        self.cotton_props = cotton_props

    def render(self, context):
        resolved_props = {}

        # if the same var is already set in context, it's being passed explicitly to override the cotton_var
        # if not, then we resolve it from the context
        for key, value in self.cotton_props.items():
            # if key in context:
            #     resolved_props[key] = context[key]
            #     continue
            try:
                resolved_props[key] = template.Variable(value).resolve(context)
            except (TypeError, template.VariableDoesNotExist):
                resolved_props[key] = value

        cotton_props = {"cotton_props": resolved_props}

        # Update the global context directly
        context.update(resolved_props)
        context.update(cotton_props)
        context.update({"cotton_props": resolved_props})
        context["cotton_props"].update(resolved_props)

        return ""



from django import template
from django.template.base import token_kwargs
from django.utils.safestring import mark_safe

register = template.Library()


def cotton_props_frame(parser, token):
    """The job of the props frame is to filter component kwargs against props. It has to be
    second component as desire to declare props inside the component template and therefore the
    component can not process context from it's own template, instead we declare the props frame
    directly inside component"""
    bits = token.split_contents()[1:]  # Skip the tag name
    # Parse token kwargs while maintaining token order
    tag_kwargs = token_kwargs(bits, parser)

    nodelist = parser.parse(("endcotton_props_frame",))
    parser.delete_first_token()
    return CottonPropsFrameNode(nodelist, tag_kwargs)


class CottonPropsFrameNode(template.Node):
    def __init__(self, nodelist, kwargs):
        self.nodelist = nodelist
        self.kwargs = kwargs

    def render(self, context):
        # Assume 'attrs' are passed from the parent and are available in the context
        parent_attrs = context.get("attrs_dict", {})

        # Initialize props based on the frame's kwargs and parent attrs
        props = {}
        for key, value in self.kwargs.items():
            # Attempt to resolve each kwarg value (which may include template variables)
            resolved_value = value.resolve(context)
            # Check if the prop exists in parent attrs; if so, use it, otherwise use the resolved default
            if key in parent_attrs:
                props[key] = parent_attrs[key]
            else:
                props[key] = resolved_value

        # Overwrite 'attrs' in the local context by excluding keys that are identified as props
        attrs_without_props = {k: v for k, v in parent_attrs.items() if k not in props}
        context["attrs_dict"] = attrs_without_props

        # Provide all of the attrs as a string to pass to the component
        def ensure_quoted(value):
            if isinstance(value, str) and value.startswith('"') and value.endswith('"'):
                return value
            else:
                return f'"{value}"'

        attrs = " ".join(
            [
                f"{key}={ensure_quoted(value)}"
                for key, value in attrs_without_props.items()
            ]
        )

        context.update({"attrs": mark_safe(attrs)})
        context.update(attrs_without_props)
        context.update(props)

        return self.nodelist.render(context)



from django.shortcuts import render

# benchmark tests


def compiled_cotton_test_view(request):
    return render(request, "compiled_cotton_test.html")


def cotton_test_view(request):
    return render(request, "cotton_test.cotton.html")


def native_extends_test_view(request):
    return render(request, "native_extends_test.html")


def native_include_test_view(request):
    return render(request, "native_include_test.html")


# Django tests


def attribute_merging_test_view(request):
    return render(request, "attribute_merging_test.cotton.html")


def attribute_passing_test_view(request):
    return render(request, "attribute_passing_test.cotton.html")


def django_syntax_decoding_test_view(request):
    return render(request, "django_syntax_decoding_test.cotton.html")
